#if ! defined SEQ66_APP_LIMITS_H
#define SEQ66_APP_LIMITS_H

/*
 *  This file is part of seq66.
 *
 *  seq66 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  seq66 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with seq66; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file          app_limits.h
 *
 *  This module holds macro constants related to the application limits of
 *  Seq66.
 *
 * \library       seq66 application
 * \author        Chris Ahlstrom
 * \date          2015-11-08
 * \updates       2021-04-22
 * \license       GNU GPLv2 or above
 *
 *  This collection of macros describes some facets of the
 *  "Patterns Panel" or "Sequences Window", which is visually presented by
 *  the Gtk::Window-derived class called mainwnd (now qsmainwnd in the Qt 5
 *  version of the application.
 *
 *  The Patterns Panel contains an 8-by-4 grid of "pattern boxes" or
 *  "sequence boxes".  All of the patterns in this grid comprise what is
 *  called a "set" (in the musical sense) or a "screen set".
 *
 *  These macros also specify other parameters, as well.
 *
 *  Why do we use macros instead of const values?  First, it really doesn't
 *  matter for simple values.  Second, we want to follow the convention that
 *  important values are all upper-case, as is convention with most
 *  macros.  They just stand out more in code. Call us old school or old
 *  fools, you decide.  Hell, we still like snprintf() for some uses!
 */

/**
 *  Default value for c_max_sets.  This is now a constant limit, as we can
 *  always support 32 sets, and have enough keystrokes for no more than that
 *  number.
 */

#define SEQ66_DEFAULT_SET_MAX             32

/**
 *  Default MIDI control input buss.  This value preserves the old behavior,
 *  where the incoming MIDI events of a device on any buss would be acted on
 *  (if specified in the MIDI control stanzas).  This value is the same as
 *  c_bussbyte_max in the midibytes.hpp module.  It can be changed in the
 *  'ctrl' file.
 */

#define SEQ66_MIDI_CONTROL_IN_BUSS        0xFF

/**
 *  Default MIDI control output buss.  It is used with igorangst's
 *  MIDI-control-out feature at present.  It can be changed in the 'ctrl'
 *  file.
 */

#define SEQ66_MIDI_CONTROL_OUT_BUSS       15

/**
 *  Maximum value for c_max_sets.
 *  Note that the largest number of sets is 4 x 8 = 32.  This limitation is
 *  necessary because there are only so many available keys on the keyboard for
 *  pattern, mute-group, and set control.
 */

#define SEQ66_MAX_SET_MAX                 32

/**
 *  Default number of rows in the main-window's grid.  This value applies to the
 *  layout of the pattern and mute-group keystrokes, as well as the virtual
 *  layout of sets into rows and columns.
 */

#define SEQ66_DEFAULT_SET_ROWS             4

/**
 *  Minimum number of rows in the main-window's grid.  This will remain
 *  the same as the default number of rows; we will not reduce the number of
 *  sequences per set, at least at this time.
 */

#define SEQ66_MIN_SET_ROWS                 4

/**
 *  Maximum number of rows in the main-window's grid.  With the default number
 *  of columns, this will double the number of sequences per set from 32 to
 *  64, hence the name "seq66".
 */

#define SEQ66_MAX_SET_ROWS                12

/**
 *  Default number of columns in the main-window's grid.
 */

#define SEQ66_DEFAULT_SET_COLUMNS          8

/**
 *  Minimum number of columns in the main-window's grid.  Currently the same
 *  as the default number.  We currently cannot support more sets than 32,
 *  which would happen if we let rows or columns go below the default 4 x 8
 *  settings.
 */

#define SEQ66_MIN_SET_COLUMNS              4

/**
 *  Maximum number of columns in the main-window's grid.  Currently the same
 *  as the default number.
 */

#define SEQ66_MAX_SET_COLUMNS             12

/**
 *  Default value for c_seqs_in_set, which is now used only in Seqtool.
 */

#define SEQ66_DEFAULT_SEQS_IN_SET \
    (SEQ66_DEFAULT_SET_ROWS * SEQ66_DEFAULT_SET_COLUMNS)

/**
 *  Default value for c_max_groups.  This value replaces c_seqs_in_set for
 *  usage in obtaining mute-group information from the "rc" file.  Its value
 *  is only "coincidentally" equal to 32.  However, given that we have limited
 *  numbers of mute-group keys, this will be the highest value that could be
 *  used.
 */

#define SEQ66_DEFAULT_GROUP_MAX           32

/**
 *  Default value of number of slot toggle keys (shortcut keys) that
 *  can be defined.  Even if we end up adding more slots to a set, this
 *  would be about the maximum number of keys we could really support.
 */

#define SEQ66_SET_KEYS_MAX                32

/**
 *  Default values of the height of the piano keys in the Qt 5 qseqkeys
 *  user-interface.  Also configurable in the "usr" file.
 */

#define SEQ66_SEQKEY_HEIGHT_MIN            6
#define SEQ66_SEQKEY_HEIGHT_DEFAULT       10
#define SEQ66_SEQKEY_HEIGHT_MAX           32    /* useful on touch-screens  */

/**
 *  The number of default virtual ALSA input busses supported in the
 *  manual-ports mode.  This value used to implicitly be 1, but it would be
 *  useful to allow a few more.  Now expanded per user request.  Let the user
 *  beware!  See issue #42.
 */

#define SEQ66_INPUT_BUSS_MAX              48
#define SEQ66_INPUT_BUSS_DEFAULT           4

/**
 *  The number of ALSA I/O busses supported.  See mastermidibus::init().
 *  Currently, this is also the default number of "manual" (virtual) output
 *  ports created in the manual-ports mode.  Now expanded per user request.
 *  Let the user beware!  See issue #42.
 */

#define SEQ66_OUTPUT_BUSS_MAX             48
#define SEQ66_OUTPUT_BUSS_DEFAULT          8

/**
 *  Flags an unspecified buss number.  Two spellings are provided, one for
 *  youngsters and one for old men.  :-D
 *
 *  Also note that the bussbyte type is an unsigned character. The
 *  midibytes.hpp file defines c_bussbyte_max as the "no bus" value, 0xFF.
 */

#define SEQ66_NO_BUS                    (0xFF)
#define SEQ66_BAD_BUS_ID                (unsigned(-1))

/**
 *  Flags an unspecified port number, or indicates a bad client ID or port
 *  number.
 */

#define SEQ66_NO_PORT                   (-1)
#define SEQ66_BAD_PORT_ID               (unsigned(-1))

/**
 *  Flags an unspecified queue number.
 */

#define SEQ66_NO_QUEUE                  (-1)
#define SEQ66_BAD_QUEUE_ID              (unsigned(-1))

/**
 *  Guessing that this has to do with the width of the performerance piano roll.
 *  See perfroll::init_before_show().
 */

#define SEQ66_PERFROLL_PAGE_FACTOR      4096

/**
 *  This constant indicates that a configuration file numeric value is
 *  the default value for specifying that an instrument is a GM
 *  instrument.  Used in the "user" configuration-file processing.
 */

#define SEQ66_GM_INSTRUMENT_FLAG          (-1)

/**
 *  This value indicates to use the default value of PPQN and ignore (to some
 *  extent) what value is specified in the MIDI file.  Note that the default
 *  default PPQN is given by the global ppqn (192) or, if the "--ppqn qn"
 *  option is specified on the command-line or the "midi_ppqn" setting in the
 *  "usr" file.
 *
 *  However, if the "midi_ppqn" setting is 0, then the default PPQN is
 *  whatever the MIDI file specifies.
 */

#define SEQ66_USE_DEFAULT_PPQN            (-1)

/**
 *  Use the PPQN from the loaded file, rather than converting to the active
 *  default PPQN of the application.
 */

#define SEQ66_USE_FILE_PPQN               0

/**
 *  Default value for the global parts-per-quarter-note value.  This is
 *  the unit of time for delta timing.  It represents the units, ticks, or
 *  pulses per beat.  Note that we're migrating this global value into the
 *  user_settings class.  Please leave this value at 192 for all eternity, for
 *  compatibility with other versions of Seq24 or Kepler34.
 */

#define SEQ66_DEFAULT_PPQN               192

/**
 *  Minimum value for PPQN.  Mostly for sanity checking.  This was set to 96,
 *  but there have been tunes set to 32 PPQN, I think.  At this setting,
 *  the Gtkmm-2.4 maintime bar moves pretty jerkily.
 */

#define SEQ66_MINIMUM_PPQN                32

/**
 *  Maximum value for PPQN.  Mostly for sanity checking, with higher values
 *  possibly useful for debugging.
 */

#define SEQ66_MAXIMUM_PPQN             19200

/**
 *  This value represent the smallest horizontal unit in a Sequencer66 grid.  It
 *  is the number of pixels in the smallest increment between vertical lines in
 *  the grid.
 */

#define SEQ66_PIXELS_PER_SUBSTEP            6

/**
 *  Minimum possible value for zoom, indicating that one pixel represents one
 *  tick.
 */

#define SEQ66_MINIMUM_ZOOM                  1

/**
 *  The default value of the zoom, indicating that one pixel represents two
 *  ticks.  However, it turns out we're going to have to support adapting the
 *  default zoom to the PPQN, in addition to allowing some extra zoom values.
 */

#define SEQ66_DEFAULT_ZOOM                  2

/**
 *  The maximum value of the zoom, indicating that one pixel represents 512
 *  ticks.  The old maximum was 32, but now that we support PPQN up to 19200,
 *  we need a couple of extra entries.
 */

#define SEQ66_MAXIMUM_ZOOM                512

/**
 *  The default value of the snap in the sequence editors.
 */

#define SEQ66_DEFAULT_SNAP                 16

/**
 *  Minimum possible value for the global redraw rate.
 */

#define SEQ66_MINIMUM_REDRAW               10

/**
 *  The default value global redraw rate.
 */

#define SEQ66_DEFAULT_REDRAW               40     /* or 25 for Windows */

/**
 *  The maximum value for the global redraw rate.
 */

#define SEQ66_MAXIMUM_REDRAW              100

/**
 *  Defines the callback rate for gtk_timeout_add() as used by perfedit.
 *  As usual, this value is in milliseconds.
 */

#define SEQ66_FF_RW_TIMEOUT               120

/**
 *  Defines a scale value for BPM so that we can store a higher-precision
 *  version of it in the proprietary "bpm" section.  See the midifile class
 *  for more information.
 */

#define SEQ66_BPM_SCALE_FACTOR          1000.0

/**
 *  The amount of time to wait for inaction before clearing the tap-button
 *  values, in milliseconds.
 */

#define SEQ66_TAP_BUTTON_TIMEOUT         5000L

/**
 *  Default value for c_beats_per_minute (global beats-per-minute, also known
 *  as "BPM").  Do not confuse this "bpm" with the other one, "beats per
 *  measure".
 */

#define SEQ66_DEFAULT_BPM                120.0

/**
 *  Minimum value for c_beats_per_minute (global beats-per-minute, also known
 *  as "BPM").  Mostly for sanity-checking, with extra low values allowed for
 *  debugging and troubleshooting.  When we have both Hydrogen and Seq66
 *  running in Slave mode, BPM alternates between 1.0 and 120.0!
 */

#define SEQ66_MINIMUM_BPM                  2.0       /* 1.0, 20.0   */

/**
 *  Maximum value for c_beats_per_minute (global beats-per-minute, also known
 *  as "BPM").  Mostly for sanity-checking.
 */

#define SEQ66_MAXIMUM_BPM                600.0       /* 500  */

/**
 *  Provides a fallback value for the BPM precision.  This is the "legacy"
 *  value.
 */

#define SEQ66_DEFAULT_BPM_PRECISION        0

/**
 *  Provides a minimum value for the BPM precision.  That is, no decimal
 *  point.
 */

#define SEQ66_MINIMUM_BPM_PRECISION        0

/**
 *  Provides a maximum value for the BPM precision, two decimal points.
 */

#define SEQ66_MAXIMUM_BPM_PRECISION        2

/**
 *  Provides a fallback value for the BPM increment.  This is the "legacy"
 *  value.
 */

#define SEQ66_DEFAULT_BPM_INCREMENT        1.0

/**
 *  Provides a minimum value for the BPM increment.
 */

#define SEQ66_MINIMUM_BPM_INCREMENT        0.01

/**
 *  Provides a maximum value for the BPM increment.
 */

#define SEQ66_MAXIMUM_BPM_INCREMENT       25.0

/**
 *  Provides a fallback value for the BPM step increment.  This is the "legacy"
 *  value.
 */

#define SEQ66_DEFAULT_BPM_STEP_INCREMENT   1.0

/**
 *  Provides a fallback value for the BPM page increment.
 */

#define SEQ66_DEFAULT_BPM_PAGE_INCREMENT  10.0

/**
 *  Minimum value for "beats-per-measure".  A new addition for the Qt 5
 *  user-interface.
 */

#define SEQ66_MINIMUM_BEATS_PER_MEASURE    1

/**
 *  Default value for "beats-per-measure".  This is the "numerator" in a 4/4
 *  time signature.  It also seems to be the value used for JACK's
 *  jack_position_t.beats_per_bar field.  For abbreviation, we will call this
 *  value "BPB", or "beats per bar", to distinguish it from "BPM", or "beats
 *  per minute".
 */

#define SEQ66_DEFAULT_BEATS_PER_MEASURE    4

/**
 *  Maximum value for "beats-per-measure".  A new addition for the Qt 5
 *  user-interface.
 */

#define SEQ66_MAXIMUM_BEATS_PER_MEASURE   32

/**
 *  Minimum number of measures in the official length of a pattern.
 */

#define SEQ66_MINIMUM_MEASURES             1

/**
 *  Default number of measures in the official length of a pattern.
 */

#define SEQ66_DEFAULT_MEASURES             1

/**
 *  Maximum number of measures in the official length of a pattern.
 */

#define SEQ66_MAXIMUM_MEASURES          1024

/**
 *  The minimum value of the beat width.  A new addition for the Qt 5
 *  user-interface.
 */

#define SEQ66_MINIMUM_BEAT_WIDTH           1

/**
 *  Default value for "beat-width".  This is the "denominator" in a 4/4 time
 *  signature.  It also seems to be the value used for JACK's
 *  jack_position_t.beat_type field. For abbreviation, we will call this value
 *  "BW", or "beat width", not to be confused with "bandwidth".
 */

#define SEQ66_DEFAULT_BEAT_WIDTH            4

/**
 *  The maximum value of the beat width.  A new addition for the Qt 5
 *  user-interface.
 */

#define SEQ66_MAXIMUM_BEAT_WIDTH           32

/**
 *  Default value for major divisions per bar.  A graphics version of
 *  SEQ66_DEFAULT_BEATS_PER_MEASURE.
 */

#define SEQ66_DEFAULT_LINES_PER_MEASURE     4

/**
 *  Default value for perfedit snap.
 */

#define SEQ66_DEFAULT_PERFEDIT_SNAP         8

/**
 *  Default value for c_thread_trigger_width_ms.
 */

#define SEQ66_DEFAULT_TRIGWIDTH_MS          4

/**
 *  Defines the minimum Note On velocity.
 */

#define SEQ66_MIN_NOTE_ON_VELOCITY          0

/**
 *  Defines the default Note On velocity, a new "stazed" feature.
 */

#define SEQ66_DEFAULT_NOTE_ON_VELOCITY    100

/**
 *  Defines the maximum Note On velocity.
 */

#define SEQ66_MAX_NOTE_ON_VELOCITY        127

/**
 *  Indicates to preserve the velocity of incoming MIDI Note events, for both
 *  on or off events.  This value represents the "Free" popup-menu entry for
 *  the "Vol" button in the seqedit window.
 */

#define SEQ66_PRESERVE_VELOCITY         (-1)

/**
 *  An older value, previously used for both Note On and Note Off velocity.
 *  See the "Stazed" note in the sequence::add_note() function.
 */

#define SEQ66_DEFAULT_NOTE_VELOCITY      100

/**
 *  Defines the default Note Off velocity, a new "stazed" feature.
 */

#define SEQ66_DEFAULT_NOTE_OFF_VELOCITY   64

/**
 *  Defines the maximum number of notes playing at one time that the
 *  application will support.
 */

#define SEQ66_PLAYING_NOTES_MAX          256

/**
 *  Provides a sanity check for transposition values.
 */

#define SEQ66_TRANSPOSE_UP_LIMIT        (128 / 2)
#define SEQ66_TRANSPOSE_DOWN_LIMIT      (-128 / 2)

#endif      // SEQ66_APP_LIMITS_H

/*
 * app_limits.h
 *
 * vim: sw=4 ts=4 wm=4 et ft=cpp
 */

