#if ! defined SEQ66_EVENT_HPP
#define SEQ66_EVENT_HPP

/*
 *  This file is part of seq66.
 *
 *  seq66 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  seq66 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with seq66; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file          event.hpp
 *
 *  This module declares/defines the event class for operating with
 *  MIDI events.
 *
 * \library       seq66 application
 * \author        Chris Ahlstrom
 * \date          2015-07-24
 * \updates       2021-02-18
 * \license       GNU GPLv2 or above
 *
 *  This module also declares/defines the various constants, status-byte
 *  values, or data values for MIDI events.  This class is also a base class,
 *  so that we can manage "editable events".
 *
 *  One thing we need to add to this event class is a way to encapsulate
 *  Meta events.  First, we use the existing event::sysex to hold
 *  this data.
 *
 *  The MIDI protocol consists of MIDI events that carry four types of messages:
 *
 *      -   Voice messages.  0x80 to 0xEF; includes channel information.
 *      -   System common messages.  0xF0 (SysEx) to 0xF7 (End of SysEx)
 *      -   System realtime messages. 0xF8 to 0xFF.
 *      -   Meta messages. 0xFF is the flag, followed by type, length, and data.
 */

#include <string>                       /* used in to_string()              */
#include <vector>                       /* SYSEX data stored in vector      */

#include "seq66_features.hpp"           /* Sequencer66 feature macros       */
#include "midi/midibytes.hpp"           /* seq66::midibyte alias, etc.      */
#include "util/basic_macros.hpp"        /* not_nullptr()                    */

/**
 *  Defines the number of data bytes in MIDI status data.
 *
 *  But consider this, events other than System Exclusive, which do not have
 *  an arbitrary number of bytes, but a definite number. These events are:
 *
\verbatim
        -   Sequence No.:   FF 00 02 s1 s1
        -   MIDI Channel:   FF 20 01 cc
        -   MIDI Port:      FF 21 01 pp
        -   Set Tempo:      FF 51 03 tt tt tt
        -   SMPTE Offset:   FF 54 05 hh mm ss fr ff
        -   Time Signature: FF 58 04 nn dd cc bb
        -   Key Signature:  FF 59 02 sf mi
\endverbatim
 *
 *  The arbitrarily-sized Meta events are:
 *
\verbatim
        -   Text:           FF 01 len text
        -   Copyright:      FF 02 len text
        -   Track Name:     FF 03 len name
        -   Instrument:     FF 04 len name
        -   Marker:         FF 05 len text
        -   Cue Point:      FF 06 len text
        -   Seq. Specific:  FF 7F len data
\endverbatim
 *
 *  The maximum amount of constant-size data is 5 bytes.  We should aim to
 *  increase this and use it, using event::m_status as the "meta" byte and
 *  perhaps m_channel as the "meta-event" byte.  But curently, the tempo and
 *  time signature events are stored as data in the sequence object, so
 *  that's probably the best tact for the future.
 */

#define SEQ66_MIDI_DATA_BYTE_COUNT      2

/*
 *  Do not document a namespace; it breaks Doxygen.
 */

namespace seq66
{

/**
 *  This highest bit of the STATUS byte is always 1.  If this bit is not set,
 *  then the MIDI byte is a DATA byte.
 */

const midibyte EVENT_STATUS_BIT         = 0x80;

/**
 *  Channel Voice Messages.
 *
 *  The following MIDI events are channel messages.  The comments represent
 *  the one or two data-bytes of the message.
 *
 *  Note that Channel Mode Messages use the same code as the Control Change,
 *  but uses reserved controller numbers ranging from 122 to 127.
 *
 *  The EVENT_ANY (0x00) value may prove to be useful in allowing any event to
 *  be dealt with.  Not sure yet, but the cost is minimal.
 */

const midibyte EVENT_ANY                = 0x00u;      // our own value
const midibyte EVENT_NOTE_OFF           = 0x80u;      // 0kkkkkkk 0vvvvvvv
const midibyte EVENT_NOTE_ON            = 0x90u;      // 0kkkkkkk 0vvvvvvv
const midibyte EVENT_AFTERTOUCH         = 0xA0u;      // 0kkkkkkk 0vvvvvvv
const midibyte EVENT_CONTROL_CHANGE     = 0xB0u;      // 0ccccccc 0vvvvvvv
const midibyte EVENT_PROGRAM_CHANGE     = 0xC0u;      // 0ppppppp
const midibyte EVENT_CHANNEL_PRESSURE   = 0xD0u;      // 0vvvvvvv
const midibyte EVENT_PITCH_WHEEL        = 0xE0u;      // 0lllllll 0mmmmmmm

/**
 *  Control Change Messages.  This is a small subset of the roughly 40 control
 *  changes.
 */

const midibyte EVENT_CTRL_VOLUME        = 0x07u;
const midibyte EVENT_CTRL_BALANCE       = 0x08u;
const midibyte EVENT_CTRL_PAN           = 0x0Au;
const midibyte EVENT_CTRL_EXPRESSION    = 0x0Bu;

/**
 *  System Messages.
 *
 *  The following MIDI events have no channel.  We have included redundant
 *  constant variables for the SysEx Start and End bytes just to make it
 *  clear that they are part of this sequence of values, though usually
 *  treated separately.
 *
 *  Only the following constants are followed by some data bytes:
 *
 *      -   EVENT_MIDI_SYSEX            = 0xF0
 *      -   EVENT_MIDI_QUARTER_FRAME    = 0xF1      // undefined?
 *      -   EVENT_MIDI_SONG_POS         = 0xF2
 *      -   EVENT_MIDI_SONG_SELECT      = 0xF3
 *
 *  A MIDI System Exclusive (SYSEX) message starts with F0, followed
 *  by the manufacturer ID (how many? bytes), a number of data bytes, and
 *  ended by an F7.
 *
 * MIDI System Real-Time Messages:
 *
 *      -   https://en.wikipedia.org/wiki/MIDI_beat_clock
 *      -   http://www.midi.org/techspecs/midimessages.php
 */

const midibyte EVENT_MIDI_REALTIME       = 0xF0u;   // 0xFn when masked
const midibyte EVENT_MIDI_SYSEX          = 0xF0u;   // redundant, see below
const midibyte EVENT_MIDI_QUARTER_FRAME  = 0xF1u;   // system common > 0 bytes
const midibyte EVENT_MIDI_SONG_POS       = 0xF2u;   // 2 data bytes
const midibyte EVENT_MIDI_SONG_SELECT    = 0xF3u;   // 1 data byte, not used
const midibyte EVENT_MIDI_SONG_F4        = 0xF4u;   // undefined
const midibyte EVENT_MIDI_SONG_F5        = 0xF5u;   // undefined
const midibyte EVENT_MIDI_TUNE_SELECT    = 0xF6u;   // 0 data bytes, not used
const midibyte EVENT_MIDI_SYSEX_END      = 0xF7u;   // redundant, see below
const midibyte EVENT_MIDI_SYSEX_CONTINUE = 0xF7u;   // redundant, see below
const midibyte EVENT_MIDI_CLOCK          = 0xF8u;   // no data bytes
const midibyte EVENT_MIDI_SONG_F9        = 0xF9u;   // undefined
const midibyte EVENT_MIDI_START          = 0xFAu;   // no data bytes
const midibyte EVENT_MIDI_CONTINUE       = 0xFBu;   // no data bytes
const midibyte EVENT_MIDI_STOP           = 0xFCu;   // no data bytes
const midibyte EVENT_MIDI_SONG_FD        = 0xFDu;   // undefined
const midibyte EVENT_MIDI_ACTIVE_SENSE   = 0xFEu;   // 0 data bytes, not used
const midibyte EVENT_MIDI_RESET          = 0xFFu;   // 0 data bytes, not used

/**
 *  0xFF is a MIDI "escape code" used in MIDI files to introduce a MIDI meta
 *  event.  Note that it has the same code (0xFF) as the Reset message, but
 *  the Meta message is read from a MIDI file, while the Reset message is sent
 *  to the sequencer by other MIDI participants.
 */

const midibyte EVENT_MIDI_META           = 0xFFu;   // an escape code

/**
 *  Provides values for the currently-supported Meta events, and many others:
 *
 *      -   Set Tempo (0x51)
 *      -   Time Signature (0x58)
 */

const midibyte EVENT_META_SEQ_NUMBER     = 0x00u;
const midibyte EVENT_META_TEXT_EVENT     = 0x01u;   // skipped
const midibyte EVENT_META_COPYRIGHT      = 0x02u;   // skipped
const midibyte EVENT_META_TRACK_NAME     = 0x03u;
const midibyte EVENT_META_INSTRUMENT     = 0x04u;   // skipped
const midibyte EVENT_META_LYRIC          = 0x05u;   // skipped
const midibyte EVENT_META_MARKER         = 0x06u;   // skipped
const midibyte EVENT_META_CUE_POINT      = 0x07u;   // skipped
const midibyte EVENT_META_MIDI_CHANNEL   = 0x20u;   // skipped, obsolete
const midibyte EVENT_META_MIDI_PORT      = 0x21u;   // skipped, obsolete
const midibyte EVENT_META_END_OF_TRACK   = 0x2Fu;
const midibyte EVENT_META_SET_TEMPO      = 0x51u;
const midibyte EVENT_META_SMPTE_OFFSET   = 0x54u;   // skipped
const midibyte EVENT_META_TIME_SIGNATURE = 0x58u;
const midibyte EVENT_META_KEY_SIGNATURE  = 0x59u;
const midibyte EVENT_META_SEQSPEC        = 0x7Fu;

/**
 *  As a "type" (overloaded on channel) value for a Meta event, 0xFF indicates
 *  an illegal meta type.
 */

const midibyte EVENT_META_ILLEGAL        = c_midibyte_max;  /* problem code */

/**
 *  These file masks are used to obtain or to mask off the channel data from a
 *  status byte.
 */

const midibyte EVENT_GET_CHAN_MASK       = 0x0Fu;
const midibyte EVENT_CLEAR_CHAN_MASK     = 0xF0u;

/**
 *  Variable from the "stazed" extras.  We reversed the parts of each token
 *  for consistency with the macros defined above.
 */

const int EVENTS_ALL                     = -1;
const int EVENTS_UNSELECTED              =  0;

/**
 *  This free function is used in the midifile module and in the
 *  event::is_note_off_recorded() member function.
 *
 * \param status
 *      The type of event, which might be EVENT_NOTE_ON.
 *
 * \param data
 *      The data byte to check.  It should be zero for a note-on is note-off
 *      event.
 */

inline bool
is_note_off_velocity (midibyte status, midibyte data)
{
    return ((status & EVENT_CLEAR_CHAN_MASK) == EVENT_NOTE_ON) && (data == 0);
}

/**
 *  Provides events for management of MIDI events.
 *
 *  A MIDI event consists of 3 bytes:
 *
 *      -#  Status byte, 1sssnnnn, where the 1sss bits specify the type of
 *          message, and the nnnn bits denote the channel number, 0 to 15.
 *          The status byte always starts with 1.
 *      -#  The first data byte, 0xxxxxxx, where the data byte always
 *          start with 0, and the xxxxxxx values range from 0 to 127.
 *      -#  The second data byte, 0xxxxxxx.
 *
 *  This class may have too many member functions.
 */

class event
{

public:

    /**
     *  Provides a type definition for a vector of midibytes.  This type will
     *  also hold the generally small amounts of data needed for Meta events,
     *  but doesn't help us encapsulate derived values, such as tempo.
     */

    using sysex = std::vector<midibyte>;

    /**
     *  The data buffer for MIDI events.  This item replaces the
     *  eventlist::Events type definition so that we can replace event
     *  pointers with iterators, safely.
     */

    using buffer = std::vector<event>;
    using iterator = buffer::iterator;
    using const_iterator = buffer::const_iterator;
    using reverse_iterator = buffer::reverse_iterator;
    using const_reverse_iterator = buffer::const_reverse_iterator;

public:

    /**
     *  Provides a key value for an event.  Its types match the
     *  m_timestamp and get_rank() function of the event class.
     */

    class key
    {

    private:

        midipulse m_timestamp;      /**< The primary key-value for the key. */
        int m_rank;                 /**< The sub-key-value for the key.     */

    public:

        key (midipulse tstamp, int rank);
        key (const event & e);
        bool operator < (const key & rhs) const;
        key (const key & ek) = default;
        key & operator = (const key & ek) = default;
    };

private:

    /**
     *  Indicates the input buss on which this event came in.  The default
     *  value is unusable: c_bussbyte_max from the midibytes.hpp module.
     */

    bussbyte m_input_buss;

    /**
     *  Provides the MIDI timestamp in ticks, otherwise known as the "pulses"
     *  in "pulses per quarter note" (PPQN).
     */

    midipulse m_timestamp;

    /**
     *  This is the status byte without the channel. The channel is included
     *  when recording MIDI, but, once a sequence with a matching channel is
     *  found, the channel nybble is cleared for storage.  The channel will be
     *  added back on the MIDI bus upon playback.  The high nybble = type of
     *  event; The low nybble = channel.  Bit 7 is present in all status
     *  bytes.
     *
     *  Note that, for status values of 0xF0 (Sysex) or 0xFF (Meta), special
     *  handling of the event can occur.  We would like to eventually use
     *  inheritance to keep the event class simple.  For now, search for
     *  "tempo" and "sysex" to tease out their implementations. Sigh.
     */

    midibyte m_status;

    /**
     *  In order to be able to handle MIDI channel-splitting of an SMF 0 file,
     *  we need to store the channel, even if we override it when playing the
     *  MIDI data.  This member adds another 4 bytes to the event object, most
     *  likely.
     *
     *  Overload:  For Meta events, where is_meta() is true, this value holds
     *  the type of Meta event. See the editable_event::sm_meta_event_names[]
     *  array.  Note that EVENT_META_ILLEGAL (0xFF) indicates an illegal Meta
     *  event.
     */

    midibyte m_channel;

    /**
     *  The two bytes of data for the MIDI event.  Remember that the
     *  most-significant bit of a data byte is always 0.  A one-byte message
     *  uses only the 0th index.
     */

    midibyte m_data[SEQ66_MIDI_DATA_BYTE_COUNT];

    /**
     *  The data buffer for SYSEX messages.  Adapted from Stazed's Seq32
     *  project on GitHub.  This object will also hold the generally small
     *  amounts of data needed for Meta events.  Compare is_sysex() to
     *  is_meta() and is_ex_data() [which tests for both].
     */

    sysex m_sysex;

    /**
     *  This event is used to link NoteOns and NoteOffs together.  The NoteOn
     *  points to the NoteOff, and the NoteOff points to the NoteOn.  See, for
     *  example, eventlist::link_notes().
     */

    iterator m_linked;

    /**
     *  Indicates that a link has been made.  This item is used [via
     *  the get_link() and link() accessors] in the sequence class.
     */

    bool m_has_link;

    /**
     *  Answers the question "is this event selected in editing."
     */

    bool m_selected;

    /**
     *  Answers the question "is this event marked in processing."  This
     *  marking is more of an internal function for purposes of reorganizing
     *  events.
     */

    bool m_marked;

    /**
     *  Answers the question "is this event being painted."  This setting is
     *  made by sequence::add_event() or add_note() if the paint parameter is
     *  true (it defaults to false).
     */

    bool m_painted;

public:

    event ();
    event (midipulse tstamp, midibyte status, midibyte d0, midibyte d1);
    event (const event & rhs);
    event & operator = (const event & rhs);
    virtual ~event ();

    /*
     * Operator overload, the only one needed for sorting events in a list
     * or a map.
     */

    bool operator < (const event & rhsevent) const;

    void set_input_bus (bussbyte b)
    {
        if (b < c_bussbyte_max)                 /* unsigned, so always > 0  */
            m_input_buss = b;
    }

    bussbyte input_bus () const
    {
        return m_input_buss;
    }

    void set_timestamp (midipulse time)
    {
        m_timestamp = time;
    }

    midipulse timestamp () const
    {
        return m_timestamp;
    }

    midibyte channel () const
    {
        return m_channel;
    }

    /**
     *  Checks the channel number to see if the event's channel matches it, or if
     *  the event has no channel.  Used in the SMF 0 track-splitting code.  The
     *  value of 0xFF is Seq66's channel value that indicates that the event's
     *  m_channel value is bogus.  However, it also means that the channel, if
     *  applicable to the event, is encoded in the m_status byte itself.  This is
     *  our work around to be able to hold a multi-channel SMF 0 track in a
     *  sequence.  In a Seq66 SMF 0 track, every event has a channel.  In a Seq66
     *  SMF 1 track, the events do not have a channel.  Instead, the channel is a
     *  global value of the sequence, and is stuffed into each event when the
     *  event is played or is written to a MIDI file.
     *
     * \param channel
     *      The channel to check.
     *
     * \return
     *      Returns true if the given channel matches the event's channel.
     */

    bool check_channel (int channel) const
    {
        return m_channel == c_midibyte_max || midibyte(channel) == m_channel;
    }

    static midibyte get_channel (midibyte m)
    {
        return m & EVENT_GET_CHAN_MASK;
    }

    static midibyte mask_status (midibyte m)
    {
        return m & EVENT_CLEAR_CHAN_MASK;
    }

    static void strip_channel (midibyte & m)
    {
        m &= EVENT_CLEAR_CHAN_MASK;
    }

    /**
     *  Static test for the status bit.
     *
     * \return
     *      Returns true if the status bit is set.  Covers 0x80 to 0xFF.
     */

    static bool is_status (midibyte m)
    {
        return (m & 0x80) != 0x00;
    }

    /**
     *  Checks for a System Category status, which is supposed to clear any
     *  running status.  We do not also allow 0xff to clear running status to
     *  prevent errors in reading a file.  An ISSUE!
     */

    static bool is_system_common (midibyte m)
    {
        return m >= EVENT_MIDI_SYSEX && m <= EVENT_MIDI_SYSEX_END;
    }

    /**
     *  Checks for a Realtime Category status, which ignores running status.
     *  Ranges from 0xF8 to 0xFF,  and m <= EVENT_MIDI_RESET is always true.
     */

    static bool is_realtime (midibyte m)
    {
        return m >= EVENT_MIDI_CLOCK;
    }

    /**
     *  Static test for the status bit.
     *
     *  CURRENTLY NOT USED ANYWHERE.
     *
     * \return
     *      Returns true if the status bit is not set.
     */

    static bool is_data (midibyte m)
    {
        return (m & EVENT_STATUS_BIT) == 0x00;
    }

    /**
     *  Static test for the channel message/statuse values: Note On, Note Off,
     *  Aftertouch, Control Change, Program Change, Channel Pressure, and
     *  Pitch Wheel.  This function no longer requires that the channel data
     *  be masked off; it is also a test for a Voice Category status.  The
     *  allowed range is 0x80 to 0xEF.
     *
     *  CURRENTLY NOT USED ANYWHERE.
     *
     * \param m
     *      The channel status or message byte to be tested, with the channel
     *      bits masked off.
     *
     *  We could add an optional boolean to cause the channel nybble to be
     *  explicitly cleared.
     *
     * \return
     *      Returns true if the byte represents a MIDI channel message.
     */

    static bool is_channel_msg (midibyte m)
    {
#if defined USE_BRUTE_FORCE
        return
        (
            m == EVENT_NOTE_ON        || m == EVENT_NOTE_OFF ||
            m == EVENT_AFTERTOUCH     || m == EVENT_CONTROL_CHANGE ||
            m == EVENT_PROGRAM_CHANGE || m == EVENT_CHANNEL_PRESSURE ||
            m == EVENT_PITCH_WHEEL
        );
#else
        return m >= EVENT_NOTE_OFF && m < EVENT_MIDI_REALTIME;
#endif
    }

    /**
     *  CURRENTLY NOT USED ANYWHERE.
     *
     * \return
     *      Returns true if the value is in the range noted above.
     */

    static bool is_voice (midibyte m)
    {
        return m >= EVENT_NOTE_OFF && m < EVENT_MIDI_REALTIME;
    }

    /**
     *  Static test for channel messages that have only one data byte: Program
     *  Change and Channel Pressure.  The rest of the channel messages have
     *  two data bytes.  This function requires that the channel data have
     *  already been masked off.
     *
     * \param m
     *      The channel status or message byte to be tested, with the channel
     *      bits masked off.
     *
     *  We could add an optional boolean to cause the channel nybble to be
     *  explicitly cleared.
     *
     * \return
     *      Returns true if the byte represents a MIDI channel message that
     *      has only one data byte.  However, if this function returns false,
     *      it might not be a channel message at all, so be careful.
     */

    static bool is_one_byte_msg (midibyte m)
    {
        return m == EVENT_PROGRAM_CHANGE || m == EVENT_CHANNEL_PRESSURE;
    }

    /**
     *  Static test for channel messages that have two data bytes: Note On,
     *  Note Off, Control Change, Aftertouch, and Pitch Wheel.  This function
     *  requires that the channel data have already been masked off.
     *
     * \param m
     *      The channel status or message byte to be tested, with the channel
     *      bits masked off.
     *
     *  We could add an optional boolean to cause the channel nybble to be
     *  explicitly cleared.
     *
     * \return
     *      Returns true if the byte represents a MIDI channel message that
     *      has two data bytes.  However, if this function returns false,
     *      it might not be a channel message at all, so be careful.
     */

    static bool is_two_byte_msg (midibyte m)
    {
#if defined USE_BRUTE_FORCE
        return
        (
            m == EVENT_NOTE_ON        || m == EVENT_NOTE_OFF   ||
            m == EVENT_CONTROL_CHANGE || m == EVENT_AFTERTOUCH ||
            m == EVENT_PITCH_WHEEL
        );
#else
        return
        (
            (m >= EVENT_NOTE_OFF && m < EVENT_PROGRAM_CHANGE) ||
            m == EVENT_PITCH_WHEEL
        );
#endif
    }

    /**
     *  Static test for a SysEx message.
     *
     * \param m
     *      The status/message byte to test, with the channel bits masked off.
     */

    static bool is_sysex_msg (midibyte m)
    {
        return m == EVENT_MIDI_SYSEX;
    }

    /**
     *  Static test for sense/reset messages.
     */

    static bool is_sense_or_reset (midibyte m)
    {
        return m == EVENT_MIDI_ACTIVE_SENSE || m == EVENT_MIDI_RESET;
    }

    /**
     *  Static test for messages that involve notes and velocity: Note On,
     *  Note Off, and Aftertouch.  This function requires that the channel
     *  nybble has already been masked off.
     *
     * \param m
     *      The channel status or message byte to be tested, with the channel
     *      bits masked off.
     *
     *  We could add an optional boolean to cause the channel nybble to be
     *  explicitly cleared.
     *
     * \return
     *      Returns true if the byte represents a MIDI note message.
     */

    static bool is_note_msg (midibyte m)
    {
        return
        (
            m == EVENT_NOTE_ON || m == EVENT_NOTE_OFF || m == EVENT_AFTERTOUCH
        );
    }

    /**
     *  Static test for messages that involve notes only: Note On and
     *  Note Off.
     *
     * \param m
     *      The channel status or message byte to be tested, with the channel
     *      bits masked off.
     *
     * \return
     *      Returns true if the byte represents a MIDI note on/off message.
     */

    static bool is_strict_note_msg (midibyte m)
    {
        return m == EVENT_NOTE_ON || m == EVENT_NOTE_OFF;
    }

    /**
     *  Static test for channel messages that are either not control-change
     *  messages, or are and match the given controller value.
     *
     * \note
     *      The old logic was the first line, but can be simplified to the
     *      second line; the third line shows the abstract representation.
     *      Also made sure of this using a couple truth tables.
     *
\verbatim
        (m != EVENT_CONTROL_CHANGE) || (m == EVENT_CONTROL_CHANGE && d == cc)
        (m != EVENT_CONTROL_CHANGE) || (d == cc)
        a || (! a && b)  =>  a || b
\endverbatim
     *
     * \param m
     *      The channel status or message byte to be tested, with the channel
     *      bits masked off.
     *
     * \param cc
     *      The desired cc value, which the datum must match, if the message is
     *      a control-change message.
     *
     * \param datum
     *      The current datum, to be compared to cc, if the message is a
     *      control-change message.
     *
     * \return
     *      Returns true if the message is not a control-change, or if it is
     *      and the cc and datum parameters match.
     */

    static inline bool is_desired_cc_or_not_cc
    (
        midibyte m, midibyte cc, midibyte datum
    )
    {
        return (m != EVENT_CONTROL_CHANGE) || (datum == cc);
    }

    /**
     *  Calculates the value of the current timestamp modulo the given
     *  parameter.
     *
     * \param modtick
     *      The tick value to mod the timestamp against.  Usually the length
     *      of the pattern receiving this event.
     */

    void mod_timestamp (midipulse modtick)
    {
        m_timestamp %= modtick;
    }

    void set_status (midibyte status);
    void set_channel_status (midibyte eventcode, midibyte channel);
    void set_meta_status (midibyte metatype);
    void set_status_keep_channel (midibyte eventcode);
    bool set_midi_event
    (
        midipulse timestamp,
        const midibyte * buffer,
        int count
    );

    /**
     *  Sets the channel "nybble", without modifying the status "nybble".
     *  It actually just sets the m_channel member.  Note that the sequence
     *  channel generally overrides this value in the usage of the event.
     *
     * \param channel
     *      The channel byte to be set.  It is masked to ensure the value
     *      ranges from 0x0 to 0xF.  This update should be safe, but we could
     *      allow c_midibyte_max if issues are uncovered.
     */

    virtual void set_channel (midibyte channel)
    {
        m_channel = (channel == c_midibyte_max) ?
            c_midibyte_max : (channel & EVENT_GET_CHAN_MASK) ;
    }

    /**
     * \getter m_status
     *      Note that we have ensured that status ranges from 0x80 to 0xFF.
     */

    midibyte get_status () const
    {
        return m_status;
    }

    /**
     *  Returns true if the event's status is *not* a control-change, but
     *  does match the given status.
     *
     * \param status
     *      The status to be checked.
     */

    bool non_cc_match (midibyte status)
    {
        return status != EVENT_CONTROL_CHANGE && m_status == status;
    }

    /**
     *  Returns true if the event's status is a control-change that matches
     *  the given status, and has a control value matching the given
     *  control-change value.
     *
     * \param st
     *      The status to be checked.
     *
     * \param cc
     *      The control-change value to be checked against the events current
     *      "d0" value.
     */

    bool cc_match (midibyte st, midibyte cc)
    {
        return st == EVENT_CONTROL_CHANGE && m_status == st && m_data[0] == cc;
    }

    /**
     *  Clears the most-significant-bit of the d0 parameter, and sets it into
     *  the first byte of m_data.  The data bytes are in a two-byte array
     *  member, m_data.  This setter is useful for Program Change and Channel
     *  Pressure events.
     *
     * \warning
     *      The second byte of m_data[] is zeroed.  If it needs to be preserved,
     *      use the two-byte get_data() and set_data() functions.
     *
     * \param d0
     *      The byte value to set as the first data byte.
     */

    void set_data (midibyte d0)
    {
        m_data[0] = d0 & 0x7F;
        m_data[1] = 0;                  /* not strictly necessary   */
    }

    /**
     *  Clears the most-significant-bit of both parameters, and sets them into
     *  the first and second bytes of m_data.
     *
     * \param d1
     *      The first byte value to set.
     *
     * \param d2
     *      The second byte value to set.
     */

    void set_data (midibyte d0, midibyte d1)
    {
        m_data[0] = d0 & 0x7F;
        m_data[1] = d1 & 0x7F;
    }

    /**
     *  Yet another overload.
     */

    void set_data (midipulse tstamp, midibyte status, midibyte d0, midibyte d1);

    /**
     *  Clears the data, useful in reusing an event to hold incoming MIDI.
     */

    void clear_data ()
    {
        m_data[0] = m_data[1] = 0;
    }

    /**
     *  Retrieves only the first data byte from m_data[] and copies it into
     *  the parameter.
     *
     * \param d0 [out]
     *      The return reference for the first byte.
     */

    void get_data (midibyte & d0) const
    {
        d0 = m_data[0];
    }

    /**
     *  Retrieves the two data bytes from m_data[] and copies each into its
     *  respective parameter.
     *
     * \param d0 [out]
     *      The return reference for the first byte.
     *
     * \param d1 [out]
     *      The return reference for the first byte.
     */

    void get_data (midibyte & d0, midibyte & d1) const
    {
        d0 = m_data[0];
        d1 = m_data[1];
    }

    /**
     *  Two alternative getters for the data bytes.  Useful for one-offs.
     */

    midibyte d0 () const
    {
        return m_data[0];
    }

    midibyte d1 () const
    {
        return m_data[1];
    }

    /**
     *  Increments the first data byte (m_data[0]) and clears the most
     *  significant bit.
     */

    void increment_data1 ()
    {
        m_data[0] = (m_data[0] + 1) & 0x7F;
    }

    /**
     *  Decrements the first data byte (m_data[0]) and clears the most
     *  significant bit.
     */

    void decrement_data1 ()
    {
        m_data[0] = (m_data[0] - 1) & 0x7F;
    }

    /**
     *  Increments the second data byte (m_data[1]) and clears the most
     *  significant bit.
     */

    void increment_data2 ()
    {
        m_data[1] = (m_data[1] + 1) & 0x7F;
    }

    /**
     *  Decrements the second data byte (m_data[1]) and clears the most
     *  significant bit.
     */

    void decrement_data2 ()
    {
        m_data[1] = (m_data[1] - 1) & 0x7F;
    }

    bool append_sysex (const midibyte * data, int len);
    bool append_sysex (midibyte data);
    bool append_meta_data (midibyte metatype, const midibyte * data, int len);
    bool append_meta_data (midibyte metatype, const std::vector<midibyte> & data);
    void restart_sysex ();

    /**
     *  Resets and adds ex data.
     *
     * \param data
     *      Provides the SysEx/Meta data.  If not provided, nothing is done,
     *      and false is returned.
     *
     * \param len
     *      The number of bytes to set.
     *
     * \return
     *      Returns true if the function succeeded.
     */

    bool set_sysex (midibyte * data, int len)
    {
        m_sysex.clear();
        return append_sysex(data, len);
    }

    sysex & get_sysex ()
    {
        return m_sysex;
    }

    const sysex & get_sysex () const
    {
        return m_sysex;
    }

    void set_sysex_size (int len)
    {
        if (len == 0)
            m_sysex.clear();
        else
            m_sysex.resize(len);
    }

    int get_sysex_size () const
    {
        return int(m_sysex.size());
    }

    /**
     *  Determines if this event is a note-on event and is not already linked.
     */

    bool on_linkable () const
    {
        return is_note_on() && ! is_linked();
    }

    /**
     *  Determines if a Note Off event is linkable to this event (which is
     *  assumed to be a Note On event).  A test used in verify_and_link().
     *
     * \param e
     *      Normally this is a Note Off event.  This status is required.
     *
     * \return
     *      Returns true if the event is a Note Off, it's the same note as
     *      this note, and the Note Off is not yet linked.
     */

    bool off_linkable (buffer::iterator & eoff) const
    {
        return
        (
            eoff->is_note_off() &&
            (eoff->get_note() == get_note()) &&
            ! eoff->is_linked()
        );
    }

    /**
     *  Sets m_has_link and sets m_link to the provided event pointer.
     *
     * \param ev
     *      Provides a pointer to the event value to set.  Since we're using
     *      an iterator, we can't use a null-pointer test that.  We assume the
     *      caller has checked that the value is not end() for the container.
     */

    void link (iterator ev)
    {
        m_linked = ev;
        m_has_link = true;
    }

    iterator link () const
    {
        return m_linked;        /* iterator could be invalid, though    */
    }

    bool is_linked () const
    {
        return m_has_link;
    }

    bool is_on_linked () const
    {
        return is_note_on() && is_linked();
    }

    void unlink ()
    {
        m_has_link = false;
    }

    void paint ()
    {
        m_painted = true;
    }

    void unpaint ()
    {
        m_painted = false;
    }

    bool is_painted () const
    {
        return m_painted;
    }

    void mark ()
    {
        m_marked = true;
    }

    void unmark ()
    {
        m_marked = false;
    }

    bool is_marked () const
    {
        return m_marked;
    }

    void select ()
    {
        m_selected = true;
    }

    void unselect ()
    {
        m_selected = false;
    }

    bool is_selected () const
    {
        return m_selected;
    }

    /**
     *  Sets m_status to EVENT_MIDI_CLOCK;
     */

    void make_clock ()
    {
        m_status = EVENT_MIDI_CLOCK;
    }

    midibyte data (int index) const    /* index not checked, for speed */
    {
        return m_data[index];
    }

    /**
     *  Assuming m_data[] holds a note, get the note number, which is in the
     *  first data byte, m_data[0].
     */

    midibyte get_note () const
    {
        return m_data[0];
    }

    /**
     *  Sets the note number, clearing off the most-significant-bit and
     *  assigning it to the first data byte, m_data[0].
     *
     * \param note
     *      Provides the note value to set.
     */

    void set_note (midibyte note)
    {
        m_data[0] = note & 0x7F;
    }

    void transpose_note (int tn);

    /**
     *  Sets the note velocity, which is held in the second data byte, and
     *  clearing off the most-significant-bit, storing it in m_data[1].
     *
     * \param vel
     *      Provides the velocity value to set.
     */

    void note_velocity (int vel)
    {
        m_data[1] = midibyte(vel) & 0x7F;
    }

    midibyte note_velocity () const
    {
        return is_note() ? m_data[1] : 0 ;
    }

    /**
     *  Check for the Note On value in m_status.  Currently assumes that the
     *  channel nybble has already been stripped.
     *
     * \return
     *      Returns true if m_status is EVENT_NOTE_ON.
     */

    bool is_note_on () const
    {
        return m_status == EVENT_NOTE_ON;
    }

    /**
     *  Check for the Note Off value in m_status.  Currently assumes that the
     *  channel nybble has already been stripped.
     *
     * \return
     *      Returns true if m_status is EVENT_NOTE_OFF.
     */

    bool is_note_off () const
    {
        return m_status == EVENT_NOTE_OFF;
    }

    /**
     *  Returns true if m_status is a Note On, Note Off, or Aftertouch message.
     *  All of these are notes, associated with a MIDI key value.  Uses the
     *  static function is_note_msg().
     *
     * \return
     *      The return value of is_note_msg() is returned.
     */

    bool is_note () const
    {
        return is_note_msg(m_status);
    }

    bool is_strict_note () const
    {
        return is_strict_note_msg(m_status);
    }

    /**
     *  Some keyboards send Note On with velocity 0 for Note Off, so we
     *  provide this function to test that during recording.
     *
     * \return
     *      Returns true if the event is a Note On event with velocity of 0.
     */

    bool is_note_off_recorded () const
    {
        return is_note_off_velocity(m_status, m_data[1]);
    }

    bool is_midi_start () const
    {
        return m_status == EVENT_MIDI_START;
    }

    bool is_midi_continue () const
    {
        return m_status == EVENT_MIDI_CONTINUE;
    }

    bool is_midi_stop () const
    {
        return m_status == EVENT_MIDI_STOP;
    }

    bool is_midi_clock () const
    {
        return m_status == EVENT_MIDI_CLOCK;
    }

    bool is_midi_song_pos () const
    {
        return m_status == EVENT_MIDI_SONG_POS;
    }

    /**
     *  Indicates if the m_status value is a one-byte message (Program Change
     *  or Channel Pressure.  Channel is stripped, because sometimes we keep
     *  the channel.
     */

    bool is_one_byte () const
    {
        return is_one_byte_msg(m_status);
    }

    /**
     *  Indicates if the m_status value is a two-byte message (everything
     *  except Program Change and Channel Pressure.  Channel is stripped,
     *  because sometimes we keep the channel.
     */

    bool is_two_bytes () const
    {
        return is_two_byte_msg(m_status);
    }

    /**
     *  Indicates if the event is a System Exclusive event or not.
     *  We're overloading the SysEx support to handle Meta events as well.
     *  Perhaps we need to split this support out at some point.
     */

    bool is_sysex () const
    {
        return m_status == EVENT_MIDI_SYSEX;
    }

    bool below_sysex () const
    {
        return m_status < EVENT_MIDI_SYSEX;
    }

    /**
     *  Indicates if the event is a Sense event or a Reset event.
     *  Currently ignored by Sequencer64.
     */

    bool is_sense_reset ()
    {
        return m_status == EVENT_MIDI_ACTIVE_SENSE ||
            m_status == EVENT_MIDI_RESET;
    }

    /**
     *  Indicates if the event is a Meta event or not.
     *  We're overloading the SysEx support to handle Meta events as well.
     */

    bool is_meta () const
    {
        return m_status == EVENT_MIDI_META;
    }

    /**
     *  Indicates if we need to use extended data (SysEx or Meta).
     *  If true, then the m_channel byte is used to encode the type of meta
     *  event.
     */

    bool is_ex_data () const
    {
        return m_status == EVENT_MIDI_META || m_status == EVENT_MIDI_SYSEX;
    }

    /**
     *  Indicates if the event is a tempo event.  See sm_meta_event_names[].
     */

    bool is_tempo () const
    {
        return is_meta() && m_channel == EVENT_META_SET_TEMPO;      /* 0x51 */
    }

    midibpm tempo () const;
    bool set_tempo (midibpm tempo);

    /**
     *  Indicates if the event is a Time Signature event.  See
     *  sm_meta_event_names[].
     */

    bool is_time_signature () const
    {
        return is_meta() && m_channel == EVENT_META_TIME_SIGNATURE; /* 0x58 */
    }

    /**
     *  Indicates if the event is a Key Signature event.  See
     *  sm_meta_event_names[].
     */

    bool is_key_signature () const
    {
        return is_meta() && m_channel == EVENT_META_KEY_SIGNATURE;  /* 0x59 */
    }

    void print () const;
    void print_note (bool showlink = true) const;
    std::string to_string () const;
    int get_rank () const;
    void rescale (int oldppqn, int newppqn);

};          // class event

/*
 * Global functions in the seq66 namespace.
 */

extern event create_tempo_event (midipulse tick, midibpm tempo);

}           // namespace seq66

#endif      // SEQ66_EVENT_HPP

/*
 * event.hpp
 *
 * vim: sw=4 ts=4 wm=4 et ft=cpp
 */

